package com.method.rscd;

import com.method.rscd.ProxyList.ProxyEndpoint;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.Proxy;
import java.net.URL;
import java.net.URLConnection;
import java.nio.file.Files;
import java.nio.file.StandardOpenOption;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import javax.xml.bind.DatatypeConverter;

public class GamepackRepository {

	private static final int[] WORLDS = { 1, 5, 7, 13, 15 };

	private static final class WorkerFactory implements ThreadFactory {

		private final AtomicInteger counter = new AtomicInteger();

		@Override
		public Thread newThread(Runnable r) {
			Thread thread = new Thread(r, "GamepackWorker-#" + counter.incrementAndGet());
			thread.setPriority(Thread.MIN_PRIORITY);
			thread.setDaemon(true);
			return thread;
		}

	}

	public static void main(String[] args) throws IOException, InterruptedException {
		ProxyList.load("proxy.txt");
		ExecutorService service = Executors.newCachedThreadPool(new WorkerFactory());
		Map<String, String> keys = new HashMap<>();
		AtomicInteger numVisited = new AtomicInteger();
		for (ProxyEndpoint proxy : ProxyList.getProxies()) {
			service.execute(() -> {
				try {
					do {
						int randomWorld = WORLDS[(int) (Math.random() * (WORLDS.length - 1))];
						Proxy httpProxy = proxy.toHttpProxy();
						String page = readWebPage(httpProxy, "http://world" + randomWorld + ".runescape.com/k=5/jav_config.ws");
						String[] lines = page.split("\\n");
						String jarName = null;
						String packName = null;
						for (String line : lines) {
							if (line.startsWith("initial_jar=")) {
								jarName = line.substring("initial_jar=".length());
							}
							if (line.matches("param=\\d+=\\b([a-zA-Z0-9-_*]{32})\\b")) {
								packName = line.substring(line.indexOf('=', 6) + 1);
							}
						}
						if (jarName == null || packName == null) {
							return;
						}
						if (packName.contains("*")) {
							// we retry if gamepack name contains a star
							continue;
						}
						numVisited.getAndIncrement();
						byte[] gamepack = downloadUrl(httpProxy, "http://world" + randomWorld + ".runescape.com/k=5/" + jarName);
						String checksum = md5(gamepack);
						if (keys.containsKey(checksum)) {
							return;
						}
						keys.put(checksum, jarName);
						writeFile(packName + ".txt", page.getBytes());
						writeFile(packName + ".jar", gamepack);
						System.out.println("Found: " + packName);
					} while (false);
				} catch (Throwable e) {
					// ignored
				}
			});
		}
		service.shutdown();
		service.awaitTermination(1, TimeUnit.HOURS);
		System.out.println("Done!");
	}

	private static void writeFile(String name, byte[] content) throws IOException {
		File file = new File("repository/" + name);
		if (!file.getParentFile().exists()) {
			file.getParentFile().mkdirs();
		}
		Files.write(file.toPath(), content, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.CREATE);
	}

	public static String readWebPage(Proxy proxy, String address) throws IOException {
		return new String(downloadUrl(proxy, address));
	}

	private static byte[] downloadUrl(Proxy proxy, String address) throws IOException {
		URL url = new URL(address);
		URLConnection connection = proxy == null ? url.openConnection() : url.openConnection(proxy);
		try (InputStream in = connection.getInputStream()) {
			try (ByteArrayOutputStream bos = new ByteArrayOutputStream()) {
				byte[] buffer = new byte[4096];
				do {
					int numRead = in.read(buffer);
					if (numRead == -1) {
						break;
					}
					bos.write(buffer, 0, numRead);
				} while (true);
				return bos.toByteArray();
			}
		}
	}

	private static String md5(byte[] data) throws NoSuchAlgorithmException {
		MessageDigest md = MessageDigest.getInstance("MD5");
		byte[] digset = md.digest(data);
		return DatatypeConverter.printHexBinary(digset);
	}

}
