package com.method.rscd.cache.store;

import java.io.EOFException;
import java.io.IOException;

/**
 * Created at: Dec 12, 2016 11:45:03 PM
 * 
 * @author Walied-Yassen A.K.A Cody
 */
public class CacheStore {

	/**
	 * The sector data buffer.
	 */
	private static final byte[] sectorBuffer = new byte[520];

	/**
	 * The index id.
	 */
	private int indexId;

	/**
	 * The data file stream.
	 */
	private BufferedDiskFile dataFile;

	/**
	 * The index file stream.
	 */
	private BufferedDiskFile indexFile;

	/**
	 * Constructs a new {@code CacheStore} object instance.
	 * 
	 * @param indexId   the index id.
	 * @param dataFile  the data file.
	 * @param indexFile the index file.
	 * @param maxSize   the maximum group size allowed in this store..
	 */
	public CacheStore(int indexId, BufferedDiskFile dataFile, BufferedDiskFile indexFile) {
		this.indexId = indexId;
		this.dataFile = dataFile;
		this.indexFile = indexFile;
	}

	/**
	 * Loads the specified group data.
	 * 
	 * @param groupId the group id.
	 * @return the loaded group data as {@code byte[]}.
	 */
	public byte[] load(int groupId) {
		synchronized (dataFile) {
			try {
				if (indexFile.size() < groupId * 6 + 6) {
					return null;
				}
				indexFile.seek(groupId * 6);
				indexFile.read(sectorBuffer, 0, 6);
				int size = (sectorBuffer[2] & 0xff) + ((sectorBuffer[1] & 0xff) << 8) + ((sectorBuffer[0] & 0xff) << 16);
				int block = ((sectorBuffer[3] & 0xff) << 16) + ((sectorBuffer[4] & 0xff) << 8) + (sectorBuffer[5] & 0xff);
				if (size < 0) {
					return null;
				}
				if (block <= 0 || block > dataFile.size() / 520L) {
					return null;
				}
				byte[] data = new byte[size];
				int offset = 0;
				int chunk = 0;
				while (offset < size) {
					if (block == 0) {
						return null;
					}
					dataFile.seek(520L * block);
					int chunkSize = size - offset;
					int myHeaderSize;
					int myGroup;
					int myChunk;
					int myNextBlock;
					int myIndex;
					if (groupId > 65535) {
						if (chunkSize > 510) {
							chunkSize = 510;
						}
						myHeaderSize = 10;
						dataFile.read(sectorBuffer, 0, chunkSize + myHeaderSize);
						myGroup = (sectorBuffer[3] & 0xff) + ((sectorBuffer[1] & 0xff) << 16) + ((sectorBuffer[0] & 0xff) << 24) + ((sectorBuffer[2] & 0xff) << 8);
						myChunk = (sectorBuffer[5] & 0xff) + ((sectorBuffer[4] & 0xff) << 8);
						myNextBlock = ((sectorBuffer[6] & 0xff) << 16) + ((sectorBuffer[7] & 0xff) << 8) + (sectorBuffer[8] & 0xff);
						myIndex = sectorBuffer[9] & 0xff;
					} else {
						if (chunkSize > 512) {
							chunkSize = 512;
						}
						myHeaderSize = 8;
						dataFile.read(sectorBuffer, 0, myHeaderSize + chunkSize);
						myGroup = ((sectorBuffer[0] & 0xff) << 8) + (sectorBuffer[1] & 0xff);
						myChunk = (sectorBuffer[3] & 0xff) + ((sectorBuffer[2] & 0xff) << 8);
						myNextBlock = ((sectorBuffer[4] & 0xff) << 16) + ((sectorBuffer[5] & 0xff) << 8) + (sectorBuffer[6] & 0xff);
						myIndex = sectorBuffer[7] & 0xff;
					}
					if (groupId != myGroup || myChunk != chunk || indexId != myIndex) {
						return null;
					}
					if (myNextBlock < 0 || myNextBlock > dataFile.size() / 520L) {
						return null;
					}
					int numRead = myHeaderSize + chunkSize;
					for (int index = myHeaderSize; index < numRead; index++) {
						data[offset++] = sectorBuffer[index];
					}
					block = myNextBlock;
					chunk++;
				}
				return data;
			} catch (IOException ioexception) {
				return null;
			}

		}
	}

	/**
	 * Stores the specified group into the store.
	 * 
	 * @param groupId the group id.
	 * @param data    the group data.
	 * @param length  the data length.
	 * @return {@code true} if it was stored successfully otherwise {@code false}.
	 */
	public boolean store(int groupId, byte[] data, int length) {
		synchronized (dataFile) {
			if (length < 0) {
				throw new IllegalArgumentException("Error group store - group size exceeds max length, group_id:" + groupId);
			}
			boolean exist = store(groupId, data, length, true);
			if (!exist) {
				exist = store(groupId, data, length, false);
			}
			if (!exist) {
				System.out.println("Failed to store [archive=" + indexId + ", group=" + groupId + "]");
			}
			return exist;
		}
	}

	/**
	 * Stores the specified group into the store.
	 * 
	 * @param groupId the group id.
	 * @param data    the group data.
	 * @param length  the data length.
	 * @param exists  does the group exist in the store already?
	 * @return {@code true} if it was stored successfully otherwise {@code false}.
	 */
	public boolean store(int groupId, byte[] data, int length, boolean exists) {
		synchronized (dataFile) {
			try {
				int block;
				if (exists) {
					if (indexFile.size() < 6 + 6 * groupId) {
						return false;
					}
					indexFile.seek(groupId * 6);
					indexFile.read(sectorBuffer, 0, 6);
					block = (sectorBuffer[5] & 0xff) + ((sectorBuffer[3] & 0xff) << 16) + ((sectorBuffer[4] & 0xff) << 8);
					if (block <= 0 || block > dataFile.size() / 520L) {
						return false;
					}
				} else {
					block = (int) ((dataFile.size() + 519L) / 520L);
					if (0 == block) {
						block = 1;
					}
				}
				sectorBuffer[0] = (byte) (length >> 16);
				sectorBuffer[1] = (byte) (length >> 8);
				sectorBuffer[2] = (byte) length;
				sectorBuffer[3] = (byte) (block >> 16);
				sectorBuffer[4] = (byte) (block >> 8);
				sectorBuffer[5] = (byte) block;
				indexFile.seek(groupId * 6);
				indexFile.write(sectorBuffer, 0, 6);
				int offset = 0;
				int chunk = 0;
				while (offset < length) {
					int myNextBlock = 0;
					if (exists) {
						dataFile.seek(520L * block);
						int myGroup;
						int myChunk;
						int myIndex;
						if (groupId > 65535) {
							try {
								dataFile.read(sectorBuffer, 0, 10);
							} catch (EOFException e) {
								break;
							}
							myGroup = (sectorBuffer[3] & 0xff) + ((sectorBuffer[0] & 0xff) << 24) + ((sectorBuffer[1] & 0xff) << 16) + ((sectorBuffer[2] & 0xff) << 8);
							myChunk = ((sectorBuffer[4] & 0xff) << 8) + (sectorBuffer[5] & 0xff);
							myNextBlock = (sectorBuffer[8] & 0xff) + ((sectorBuffer[7] & 0xff) << 8) + ((sectorBuffer[6] & 0xff) << 16);
							myIndex = sectorBuffer[9] & 0xff;
						} else {
							try {
								dataFile.read(sectorBuffer, 0, 8);
							} catch (EOFException r) {
								break;
							}
							myGroup = (sectorBuffer[1] & 0xff) + ((sectorBuffer[0] & 0xff) << 8);
							myChunk = (sectorBuffer[3] & 0xff) + ((sectorBuffer[2] & 0xff) << 8);
							myNextBlock = (sectorBuffer[6] & 0xff) + ((sectorBuffer[4] & 0xff) << 16) + ((sectorBuffer[5] & 0xff) << 8);
							myIndex = sectorBuffer[7] & 0xff;
						}
						if (groupId != myGroup || chunk != myChunk || myIndex != indexId) {
							return false;
						}
						if (myNextBlock < 0 || myNextBlock > dataFile.size() / 520L) {
							return false;
						}
					}
					if (myNextBlock == 0) {
						exists = false;
						myNextBlock = (int) ((dataFile.size() + 519L) / 520L);
						if (myNextBlock == 0) {
							myNextBlock++;
						}
						if (myNextBlock == block) {
							myNextBlock++;
						}
					}
					if (groupId > 65535) {
						if (length - offset <= 510) {
							myNextBlock = 0;
						}
						sectorBuffer[0] = (byte) (groupId >> 24);
						sectorBuffer[1] = (byte) (groupId >> 16);
						sectorBuffer[2] = (byte) (groupId >> 8);
						sectorBuffer[3] = (byte) groupId;
						sectorBuffer[4] = (byte) (chunk >> 8);
						sectorBuffer[5] = (byte) chunk;
						sectorBuffer[6] = (byte) (myNextBlock >> 16);
						sectorBuffer[7] = (byte) (myNextBlock >> 8);
						sectorBuffer[8] = (byte) myNextBlock;
						sectorBuffer[9] = (byte) indexId;
						dataFile.seek(block * 520L);
						dataFile.write(sectorBuffer, 0, 10);
						int numWrite = length - offset;
						if (numWrite > 510) {
							numWrite = 510;
						}
						dataFile.write(data, offset, numWrite);
						offset += numWrite;
					} else {
						if (length - offset <= 512) {
							myNextBlock = 0;
						}
						sectorBuffer[0] = (byte) (groupId >> 8);
						sectorBuffer[1] = (byte) groupId;
						sectorBuffer[2] = (byte) (chunk >> 8);
						sectorBuffer[3] = (byte) chunk;
						sectorBuffer[4] = (byte) (myNextBlock >> 16);
						sectorBuffer[5] = (byte) (myNextBlock >> 8);
						sectorBuffer[6] = (byte) myNextBlock;
						sectorBuffer[7] = (byte) indexId;
						dataFile.seek(520L * block);
						dataFile.write(sectorBuffer, 0, 8);
						int numWrite = length - offset;
						if (numWrite > 512) {
							numWrite = 512;
						}
						dataFile.write(data, offset, numWrite);
						offset += numWrite;
					}
					block = myNextBlock;
					chunk++;
				}
				return true;
			} catch (IOException ioexception) {
				return false;
			}
		}
	}

	/**
	 * Gets the number of files stored in this file store.
	 * 
	 * @return This file store's file count.
	 */
	public int getFileCount() {
		return (int) (indexFile.size() / 6);
	}
}
